/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc. Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * Copyright (C) 2023 Daten-Verarbeitung Tirol GmbH.
 * All rights reserved.
 *
 * License: GPL (version 3 or any later version).
 * See LICENSE for details.
 * END COPYRIGHT BLOCK **/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


/************************************************************

 hibpcheck.c

 This source file provides a post-operation plug-in function that forwards
 hashes to the DVT HIBP API. This program is based on 389-ds-base-1.4.3.

 To install this plug-in function, stop the server, edit the dse.ldif file
 (in the config directory)
 and add the following lines before restarting the server :

 dn: cn=HIBP,cn=plugins,cn=config
 objectClass: top
 objectClass: nsSlapdPlugin
 objectClass: extensibleObject
 cn: HIBP
 nsslapd-pluginPath: <server_root>/plugins/slapd/slapi/examples/libhibp-plugin.so
 nsslapd-pluginInitfunc: hibp_init
 nsslapd-pluginType: postoperation
 nsslapd-pluginEnabled: on
 nsslapd-plugin-depends-on-type: database
 nsslapd-pluginId: hibp

 ************************************************************/
#include <stdio.h>
#include <string.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <slapi-plugin.h>  // dirsrv
#include <pk11func.h> // nss
#include <pk11pqg.h> // nss


#define VENDOR "DVT"
#define DS_PACKAGE_VERSION "1.0.0"
#define OUT_SIZE 2048
#define LISTEN_ADDR "127.0.0.1"
#define LISTEN_PORT 508 // an (unassigned) port in the well-known range, so only root can listen on it

Slapi_PluginDesc bindpdesc = {"hibp-check", VENDOR, DS_PACKAGE_VERSION,
                              "checks hashes against DVT HIBP API"};

int socket_desc;
struct sockaddr_in server_addr;

int hibp(Slapi_PBlock *pb) {
    char *dn;
    ber_tag_t method;  // https://lists.fedoraproject.org/archives/list/389-users@lists.fedoraproject.org/message/7FFE2HJ5IIOZ5F4YUUVTRNRGFH4T7FBE/
    struct berval *credentials;
    int status; // get auth status from https://pagure.io/389-ds-base/blob/master/f/ldap/servers/slapd/bind.c#_860
    unsigned char hash_out[SHA1_LENGTH];
    char buf[OUT_SIZE], *b = buf;
    int dn_len = 0;

    /* Gets parameters available when processing an LDAP bind operation. */
    if (slapi_pblock_get(pb, SLAPI_BIND_TARGET, &dn) != 0 ||
        slapi_pblock_get(pb, SLAPI_BIND_METHOD, &method) != 0 ||
        slapi_pblock_get(pb, SLAPI_PLUGIN_OPRETURN, &status) != 0 ||
        slapi_pblock_get(pb, SLAPI_BIND_CREDENTIALS, &credentials) != 0) {

        slapi_log_error(SLAPI_LOG_PLUGIN, "hibp", "Could not get parameters for bind operation\n");
        return 0;
    }
    if (method != LDAP_AUTH_SIMPLE) { //case LDAP_AUTH_NONE: case LDAP_AUTH_SASL:
        slapi_log_error(SLAPI_LOG_PLUGIN, "hibp", "Unsupported authentication method requested: %d\n", method);
        return 0;
    }
    if (status != SLAPI_BIND_SUCCESS) { // https://pagure.io/389-ds-base/blob/master/f/ldap/servers/slapd/slapi-plugin.h#_400
        slapi_log_error(SLAPI_LOG_PLUGIN, "hibp", "Bind was anon or did not succeed: %d\n", status);
        return 0;
    }

    /* hash password as sha1 (raw bytes returned) using libnss */
    PK11_HashBuf(SEC_OID_SHA1, hash_out, (unsigned char *)credentials->bv_val, strlen(credentials->bv_val));

    /* append dn to hash, plus newline, plus null terminator */
    for (size_t i = 0;  i < SHA1_LENGTH;  i++) {
        b += snprintf(b, OUT_SIZE-(b-buf), "%.2x", hash_out[i]);
    }
    b += snprintf(b, OUT_SIZE-(b-buf), "\t%s\n", dn);

    sendto(socket_desc, buf, strlen(buf), 0, (struct sockaddr*)&server_addr, sizeof(server_addr));

    return 0;
}

int hibp_init(Slapi_PBlock *pb) {

    slapi_log_error(SLAPI_LOG_PLUGIN, "hibp", "hibp_init called.\n");

    /* Register the post-operation bind function and specify the server plug-in version. */
    if (slapi_pblock_set(pb, SLAPI_PLUGIN_VERSION, SLAPI_PLUGIN_VERSION_01) != 0 ||
        slapi_pblock_set(pb, SLAPI_PLUGIN_DESCRIPTION, (void *)&bindpdesc) != 0 ||
        slapi_pblock_set(pb, SLAPI_PLUGIN_POST_BIND_FN, (void *)hibp) != 0) {

        slapi_log_error(SLAPI_LOG_PLUGIN, "hibp_init", "Failed to set version and functions\n");
        return -1;
    }

    socket_desc = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    server_addr.sin_family = AF_INET;
    server_addr.sin_port = htons(LISTEN_PORT);
    server_addr.sin_addr.s_addr = inet_addr(LISTEN_ADDR);

    return 0;
}
